from datetime import date, timedelta

import pytest

from pyjpboatrace.const import NUM_RACES, NUM_STADIUMS
from pyjpboatrace.validator import (
    validate_date,
    validate_race,
    validate_stadium,
)


def test_validate_date_without_value_error():
    validate_date(date.today())
    assert True


def test_validate_date_with_value_error():
    d = date.today() + timedelta(days=1)
    with pytest.raises(ValueError):
        validate_date(d)


def test_validate_stadium_without_value_error():
    for i in range(1, NUM_STADIUMS+1):
        validate_stadium(i)
    assert True


@pytest.mark.parametrize(
    "stadium",
    [
        0,
        NUM_STADIUMS+1,
    ]
)
def test_validate_stadium_with_value_error(stadium: int):
    with pytest.raises(ValueError):
        validate_stadium(stadium)


def test_validate_race_without_value_error():
    for r in range(1, NUM_RACES+1):
        validate_race(r)
    assert True


@pytest.mark.parametrize(
    "race",
    [
        0,
        NUM_RACES+1,
    ]
)
def test_validate_race_with_value_error(race: int):
    with pytest.raises(ValueError):
        validate_race(race)



from bs4 import BeautifulSoup

from ...const import BOATS_GEN
from ...exceptions import NoDataException, RaceCancelledException, UnexpectedException
from ...utils import str2num


def parse_html_raceresult(html: str):

    def parse_ranks(tbodys):
        # preparation
        lst_tds = list(
            map(lambda tbody: tbody.select('tbody > tr > td'), tbodys))
        return [
            {
                'rank': str2num(tds[0].text, int, tds[0].text),
                'boat': str2num(tds[1].text, int, ''),
                'racerid': str2num(tds[2].select('span')[0].text, int, ''),
                'name': ''.join(tds[2].select('span')[1].text.split()),
                'time': ''.join(tds[3].text.split()),
            }
            for tds in lst_tds
        ]

    def parse_starts(tds):
        # preparation
        dic = {f'course{i}': {} for i in BOATS_GEN}

        # extract
        for i, td in enumerate(tds):
            dic[f'course{i+1}']['boat'] = str2num(
                td.select('span')[0].text, int, '')
            dic[f'course{i+1}']['ST'] = str2num(
                td.select('span')[-1].text.split()[0].replace('F.', '-0.'),
                float,
                ''
            )
        return dic

    def parse_payoff(tbodys):
        # preparation
        dic = {}
        lst_trs = list(map(lambda tbody: tbody.select('tbody > tr'), tbodys))

        def extract_result(td):
            try:
                return td.select_one(
                    'div.numberSet1 > div.numberSet1_row'
                ).text.split()
            except AttributeError:
                if "不成立" in td.text:
                    return ["不成立"]
                return [f"FailedToParse: {td.text}"]

        # trifecta
        trs = lst_trs[0]
        dic['trifecta_all'] = [
            {
                'result': ''.join(tr.select('td')[-3].text.split()),
                'payoff': str2num(
                    tr.select('td')[-2]
                      .text
                      .replace('¥', '')
                      .replace(',', ''),
                    int, ''
                ),
                'popularity': str2num(tr.select('td')[-1].text, int, '')
            }
            for tr in trs
        ]
        dic['trifecta_all'] = [
            d for d in dic['trifecta_all']
            if any(map(lambda v: v != '', d.values()))
        ]
        # trio
        trs = lst_trs[1]
        dic['trio_all'] = [
            {
                'result': ''.join(tr.select('td')[-3].text.split()),
                'payoff': str2num(
                    tr.select('td')[-2].text.replace('¥', '').replace(',', ''),
                    int, ''
                ),
                'popularity': str2num(tr.select('td')[-1].text, int, '')
            }
            for tr in trs
        ]
        dic['trio_all'] = [
            d for d in dic['trio_all']
            if any(map(lambda v: v != '', d.values()))
        ]
        # exacta
        trs = lst_trs[2]
        dic['exacta_all'] = [
            {
                'result': ''.join(tr.select('td')[-3].text.split()),
                'payoff': str2num(
                    tr.select('td')[-2]
                      .text
                      .replace('¥', '')
                      .replace(',', ''),
                    int, ''
                ),
                'popularity': str2num(tr.select('td')[-1].text, int, '')
            }
            for tr in trs
        ]
        dic['exacta_all'] = [
            d for d in dic['exacta_all']
            if any(map(lambda v: v != '', d.values()))
        ]
        # quinella
        trs = lst_trs[3]
        dic['quinella_all'] = [
            {
                'result': ''.join(tr.select('td')[-3].text.split()),
                'payoff': str2num(
                    tr.select('td')[-2]
                      .text
                      .replace('¥', '')
                      .replace(',', ''),
                    int, ''
                ),
                'popularity': str2num(tr.select('td')[-1].text, int, '')
            }
            for tr in trs
        ]
        dic['quinella_all'] = [
            d for d in dic['quinella_all']
            if any(map(lambda v: v != '', d.values()))
        ]
        # quinella place
        trs = lst_trs[4]
        dic['quinella_place_all'] = [
            {
                'result': ''.join(tr.select('td')[-3].text.split()),
                'payoff': str2num(
                    tr.select('td')[-2]
                      .text
                      .replace('¥', '')
                      .replace(',', ''),
                    int,
                    ''
                ),
                'popularity': str2num(tr.select('td')[-1].text, int, '')
            }
            for tr in trs
        ]
        dic['quinella_place_all'] = [
            d for d in dic['quinella_place_all']
            if any(map(lambda v: v != '', d.values()))
        ]
        # win
        trs = lst_trs[5]
        dic['win_all'] = [
            {
                'result': ''.join(tr.select('td')[-3].text.split()),
                'payoff': str2num(
                    tr.select('td')[-2]
                      .text
                      .replace('¥', '')
                      .replace(',', ''),
                    int, ''
                ),
                'popularity': str2num(tr.select('td')[-1].text, int, '')
            }
            for tr in trs
        ]
        dic['win_all'] = [
            d for d in dic['win_all']
            if any(map(lambda v: v != '', d.values()))
        ]
        # place show
        trs = lst_trs[6]
        dic['place_show_all'] = [
            {
                'result': ''.join(tr.select('td')[-3].text.split()),
                'payoff': str2num(
                    tr.select('td')[-2]
                      .text
                      .replace('¥', '')
                      .replace(',', ''),
                    int, ''
                ),
                'popularity': str2num(tr.select('td')[-1].text, int, '')
            }
            for tr in trs
        ]
        dic['place_show_all'] = [
            d for d in dic['place_show_all']
            if any(map(lambda v: v != '', d.values()))
        ]

        # Code to ensure backward compatibility
        # trifecta
        dic['trifecta'] = dic['trifecta_all'][0]
        # trio
        dic['trio'] = dic['trio_all'][0]
        # exacta
        dic['exacta'] = dic['exacta_all'][0]
        # quinella
        dic['quinella'] = dic['quinella_all'][0]
        # quinella place
        dic['quinella_place'] = dic['quinella_place_all']
        # win
        dic['win'] = dic['win_all'][0]
        # place show
        dic['place_show'] = dic['place_show_all']

        return dic

    def parse_weather(divs):
        # direction and temperature
        div = divs[0]
        p = [
            c.replace('is-direction', '')
            for c in div.select_one('p')['class'] if 'is-direction' in c
        ][0]
        direction = str2num(p, int, '')
        span = div.select('div.weather1_bodyUnitLabel > span')[-1]
        tempereture = str2num(''.join(span.text.split())[:-1], float, '')
        # weather
        div = divs[1]
        weather = ''.join(div.select_one(
            'div.weather1_bodyUnitLabel > span').text.split())
        # wind speed
        div = divs[2]
        wind_speed = str2num(
            div.select_one(
                'div.weather1_bodyUnitLabel > span.weather1_bodyUnitLabelData'
            ).text.replace('m', ''),
            int,
            ''
        )
        # wind direction
        div = divs[3]
        p = [
            c.replace('is-wind', '')
            for c in div.select_one('p')['class'] if 'is-wind' in c
        ][0]
        wind_direction = str2num(p, int, '')
        # water temperature
        div = divs[4]
        span = div.select('div.weather1_bodyUnitLabel > span')[-1]
        water_tempereture = str2num(''.join(span.text.split())[:-1], float, '')
        # wave height
        div = divs[5]
        wave_height = str2num(
            div.select_one(
                'div.weather1_bodyUnitLabel > span.weather1_bodyUnitLabelData'
            ).text.replace('cm', ''),
            int,
            ''
        )
        return {
            'direction': direction,
            'temperature': tempereture,
            'weather': weather,
            'wind_speed': wind_speed,
            'wind_direction': wind_direction,
            'water_temperature': water_tempereture,
            'wave_height': wave_height,
        }

    def parse_returns(divs):
        lst = sum(
            map(
                lambda div: list(map(lambda span: span.text, div.select('span'))),  # noqa
                divs
            ),
            []
        )
        lst = list(map(lambda s: str2num(s, int, ''), lst))
        return [e for e in lst if e]

    def parse_kimarite(td):
        return ''.join(td.text.split())

    def parse_note(tds):
        lst = list(map(lambda td: ''.join(td.text.split()), tds))
        return [e for e in lst if e]

    # make soup
    soup = BeautifulSoup(html, 'html.parser')

    # check cancel
    if 'レース中止' in soup.text:
        raise RaceCancelledException()

    # check data
    if '※ データはありません。' in map(
        lambda e: e.text,
        soup.select('h3.title12_title')
    ):
        raise NoDataException()

    # table
    grid_units = soup.select('div.grid.is-type2.h-clear > div.grid_unit')  # probably 4 units # noqa
    if len(grid_units) < 4:
        raise NoDataException(
            "Race result data is not yet fully available. "
            "Please wait a moment and try again. "
            "If the issue persists, please report it as a bug at "
            "https://github.com/hmasdev/pyjpboatrace/issues"
        )
    ranks_table = grid_units[0].select('div.table1 > table > tbody')
    starts_table = grid_units[1].select('div.table1 > table > tbody > tr > td')
    payoff_table = grid_units[2].select('div.table1 > table > tbody')

    inner_grid_units = grid_units[3].select(
        'div.grid.is-type6.h-clear > div.grid_unit'
    )  # probably 2 units

    if len(inner_grid_units) < 2:
        raise UnexpectedException(
            "Unexpected HTML structure encountered while parsing race results. "
            "Please report this as a bug at "
            "https://github.com/hmasdev/pyjpboatrace/issues"
        )

    weather_table = inner_grid_units[0].select(
        'div.weather1 > div.weather1_body > div.weather1_bodyUnit'
    )

    return_table = inner_grid_units[1].select('div.table1')[0].select(
        'table > tbody > tr > td > div.numberSet1 >div.numberSet1_row'
    )

    kimarite = inner_grid_units[1].select('div.table1')[1]\
        .select_one('table > tbody > tr > td')

    note_table = grid_units[3].select('div.table1')[-1]\
        .select('table > tbody > tr > td')

    # parse
    dic = {
        'result': parse_ranks(ranks_table),
        'start_information': parse_starts(starts_table),
        'payoff': parse_payoff(payoff_table),
        'weather_information': parse_weather(weather_table),
        'return': parse_returns(return_table),
        'kimarite': parse_kimarite(kimarite),
        'note': parse_note(note_table)
    }

    return dic



import pytest

from pyjpboatrace.certification import check_login_status, login, logout
from pyjpboatrace.exceptions import LoginFailException
from pyjpboatrace.user_information import UserInformation

from ._driver_fixutures import driver_not_http_get_driver  # noqa
from ._utils import get_user_info


@pytest.mark.skipif(
    get_user_info() is None,
    reason='UserInformation not found'
)
@pytest.mark.integrate
def test_login_logout(driver_not_http_get_driver):  # noqa
    # preparation
    driver = driver_not_http_get_driver
    user = get_user_info()
    # pre-status
    assert not check_login_status(driver)
    # login
    assert login(driver, user)
    assert check_login_status(driver)
    # logout
    assert logout(driver)
    assert not check_login_status(driver)


@pytest.mark.integrate
def test_login_wrong_pass(driver_not_http_get_driver):  # noqa
    # preparation
    driver = driver_not_http_get_driver
    user = UserInformation('a', 'b', 'c', 'd')
    # pre-status
    assert not check_login_status(driver)
    # login
    with pytest.raises(LoginFailException):
        login(driver, user)
    assert not check_login_status(driver)



from bs4 import BeautifulSoup

from ...exceptions import RaceCancelledException
from ...utils import str2num
from .scrape_odds_update_time import scrape_odds_update_time


def parse_html_odds2tf(html: str):

    def parse_exacta(trs):
        # preparation
        lst_tds = [tr.select('td') for tr in trs]
        # extract
        dic = {}
        for i, tds in enumerate(lst_tds):
            b2 = i + 2
            for j, tde in enumerate(tds[1::2]):
                b1 = j + 1
                dic[f'{b1}-{b2}'] = str2num(tde.text, float, tde.text)
                if i == j:  # upper right of table
                    b2 = b2 - 1

        return dic

    def parse_quinella(trs):
        # preparation
        lst_tds = [tr.select('td') for tr in trs]
        # extract
        dic = {}
        for i, tds in enumerate(lst_tds):
            b2 = i + 2
            for j, tde in enumerate(tds[1::2]):
                b1 = j + 1
                if b1 >= b2:  # invalid quinella
                    break
                dic[f'{b1}={b2}'] = str2num(tde.text, float, tde.text)

        return dic

    # make soup
    soup = BeautifulSoup(html, 'html.parser')

    # check cancel
    if '※ 該当レースは中止になりました。' in soup.text:
        raise RaceCancelledException()

    # table
    tables = soup.select('div.table1')  # probably 3 tables
    exacta_table = tables[-2].select('table > tbody > tr')
    quinella_table = tables[-1].select('table > tbody > tr')

    # parse
    dic = {
        'update': scrape_odds_update_time(soup),
        'exacta': parse_exacta(exacta_table),
        'quinella': parse_quinella(quinella_table),
    }

    return dic



from bs4 import BeautifulSoup

from ...exceptions import RaceCancelledException
from ...utils import str2num
from .scrape_odds_update_time import scrape_odds_update_time


def parse_html_oddsk(html: str):

    def parse_quinellaplace(trs):

        dic = {}
        for i, tr in enumerate(trs):
            tds = tr.select('td')
            b2 = i + 2
            for j, (_, tde) in enumerate(zip(tds[::2], tds[1::2])):
                b1 = j + 1
                if b1 >= b2:  # invalid quinella-place
                    break
                dic[f'{b1}={b2}'] = list(map(
                    lambda s: str2num(s, float, s),
                    tde.text.split('-')
                ))

        return dic

    # make soup
    soup = BeautifulSoup(html, 'html.parser')

    # check cancel
    if '※ 該当レースは中止になりました。' in soup.text:
        raise RaceCancelledException()

    # table
    tables = soup.select('div.table1')  # probably 2 tables
    quinellaplace_table = tables[-1].select('table > tbody > tr')

    # parse
    dic = parse_quinellaplace(quinellaplace_table)
    dic["update"] = scrape_odds_update_time(soup)

    return dic



import datetime

from bs4 import BeautifulSoup

grade_mapper = {
    'is-ippan': 'ippan',
    'is-G3b': 'G3',
    'is-G2b': 'G2',
    'is-G1b': 'G1',
    'is-G3a': 'G3',
    'is-G2a': 'G2',
    'is-G1a': 'G1',
    'is-SGa': "SG",
    'is-lady': 'lady',
    'is-venus': 'venus',
    'is-rookie__3rdadd': 'rookie',
}

timeframe_mapper = {
    'is-morning': 'morning',
    'is_morning': 'morning',
    'is_summer': 'summer',
    'is-summer': 'summer',
    'is_nighter': 'nighter',
    'is-nighter': 'nighter',
    'is_midnight': 'midnight',
    'is-midnight': 'midnight',
}


def parse_html_index(html: str):

    # make table row parser
    def parse_trs(trs):
        # extract tds
        tds = trs[0].select('td')[::-1]

        # extract values
        stadium = tds.pop().select_one('a > img')['alt'].replace('>', '')
        status = tds.pop()

        if not status.has_attr('colspan'):
            # case : there is a race which has not started yet.
            next_race = int(tds.pop().get_text().replace('R', ''))
            next_vote_limit = trs[1].select_one('td').get_text()
            tds.pop()  # ignore vote button
        elif "前日" in status.text:
            # case : pre-sales is ongoing or completed
            tds.pop()  # ignore vote button

        grade = tds.pop()['class']
        timeframe = tds.pop().get("class", "")
        title = tds.pop().select_one('a').get_text()
        period, day = tds.pop().get_text('\n').split('\n')

        # clean values
        # grade
        grade = list(map(
            lambda s: grade_mapper.get(s, '[FailedToParse]'+s),
            grade
        ))
        # status
        status = status.get_text()
        # time frame
        timeframe = list(map(
            lambda s: timeframe_mapper.get(s, '[FailedToParse]'+s),
            timeframe
        ))
        timeframe = timeframe[0] if timeframe else ''
        # period
        period = period.split('-')

        # to dict
        ldic = {
            'status': status,
            'grade': grade,
            'timeframe': timeframe,
            'title': title,
            'period': period,
            'day': day,
        }

        if 'next_race' in locals():
            ldic['next_race'] = next_race

        if 'next_vote_limit' in locals():
            ldic['next_vote_limit'] = next_vote_limit

        return stadium, ldic

    def normalize_next_vote_limit(ldic):
        if 'next_vote_limit' in ldic:
            temp = ' '.join([d.strftime('%Y-%m-%d'), ldic['next_vote_limit']])
            temp = datetime.datetime.strptime(temp, '%Y-%m-%d %H:%M')
            ldic['next_vote_limit'] = temp.strftime('%Y-%m-%d %H:%M:%S')

    def normalize_period(ldic):
        # to datetime
        ldic['period'] = list(map(
            lambda md: datetime.datetime.strptime(
                '/'.join([str(d.year), md]), '%Y/%m/%d'),
            ldic['period']
        ))
        # validation
        if ldic['period'][0] > ldic['period'][1]:
            ldic['period'][0] = datetime.datetime(
                ldic['period'][0].year-1,
                ldic['period'][0].month,
                ldic['period'][0].day
            )
        ldic['period'] = list(
            map(lambda dt: dt.strftime('%Y-%m-%d'), ldic['period']))

    # make soup
    soup = BeautifulSoup(html, 'html.parser')

    # extract date
    d = datetime.datetime.strptime(
        soup.select_one('p.btnGroup3_refreshBtn > a')['href'].split('hd=')[-1],  # type: ignore  # noqa
        '%Y%m%d'
    )

    # table
    table = soup.select_one('div.table1').select('table > tbody')  # type: ignore  # noqa

    # parse
    dic = dict([
        parse_trs(tbody.select('tr'))
        for tbody in table
    ])

    # normalize
    for v in dic.values():
        normalize_next_vote_limit(v)
        normalize_period(v)

    return dic



from typing import Callable

import requests
from requests import Response
from requests.exceptions import ConnectionError, InvalidSchema
from selenium import webdriver
from selenium.common.exceptions import InvalidArgumentException, WebDriverException
from selenium.webdriver.edge.options import Options as EdgeOptions


def create_chrome_driver() -> webdriver.Chrome:
    """Create an instance of chrome driver with headless mode

    Returns:
        webdriver.Chrome: chrome driver
    """
    # options
    options = webdriver.ChromeOptions()
    options.add_argument('--headless')
    options.add_argument('--disable-gpu')
    options.add_argument('--ignore-certificate-errors')
    options.add_argument('--allow-running-insecure-content')
    options.add_argument('--disable-web-security')
    options.add_argument('--lang=ja-JP,ja;q=0.9,en-US;q=0.8,en;q=0.7')
    options.add_argument('--blink-settings=imagesEnabled=false')
    options.add_argument(
        '--user-agent='
        '"Mozilla\\/5.0 (Windows NT 10.0; Win64; x64) '
        'AppleWebKit\\/537.36 (KHTML, like Gecko) '
        'Chrome\\/89.0.4389.90 '
        '\\Safari/537.36"'
    )
    # create driver
    driver = webdriver.Chrome(options=options)
    return driver


def create_firefox_driver() -> webdriver.Firefox:
    """Create an instance of firefox driver with headless mode

    Returns:
        webdriver.Firefox: firefox driver
    """
    # options
    options = webdriver.FirefoxOptions()
    options.add_argument('-headless')
    # profile
    profile = webdriver.FirefoxProfile()
    profile.set_preference(
        'general.useragent.override',
        ''.join([
            '"Mozilla/5.0 (Windows NT 10.0; Win64; x64) ',
            'AppleWebKit/537.36 (KHTML, like Gecko) ',
            'Chrome/89.0.4389.90 ',
            'Safari/537.36"',
        ])
    )
    options.profile = profile
    # create driver
    driver = webdriver.Firefox(options=options)
    return driver


def create_edge_driver() -> webdriver.Edge:
    """Create an instance of edge driver

    Returns:
        webdriver.Edge: chrome driver

    WARNING:
        driver is activated without headless mode
    """
    # options
    options = EdgeOptions()
    options.add_argument("--headless=new")  # TODO set user agent
    options.add_argument("disable-gpu")
    # create driver
    driver = webdriver.Edge(options=options)
    return driver


class HTTPGetDriver:
    """
    Wrapper class for requests.get as a webdriver of selenium,
    altough this class provides only get method and page_source property.
    """

    # TODO Need inheriting selenium.webdriver.remote.webdriver.BaseWebDriver

    def __init__(
        self,
        http_get: Callable[[str, ], Response],
    ):
        self.__get = http_get
        self.__page_source = ''

    def get(self, url: str):
        """
        Loads a web page in the current browser session.
        """
        try:
            self.__page_source = self.__get(url).text
        except ConnectionError as e:
            raise WebDriverException from e
        except InvalidSchema as e:
            raise InvalidArgumentException from e
        return None

    @property
    def page_source(self):
        """
        Gets the source of the current page.
        :Usage:
            ::
                driver.page_source
        """
        return self.__page_source

    def close(self):
        pass


def create_httpget_driver(
    http_get: Callable[[str, ], Response] = requests.get,
) -> HTTPGetDriver:
    """Craete an instance of HTTPGetDriver

    Args:
        http_get (Callable[[str, ], Response], optional): HTTP get function.
            Defaults to requests.get.

    Returns:
        HTTPGetDriver: [description]
    """
    return HTTPGetDriver(http_get)



import datetime
from logging import Logger, getLogger
from typing import Any, Dict

from selenium import webdriver

from ..const import BOATRACEJP_BASE_URL
from ..validator import validate_date, validate_race, validate_stadium
from ._parser import parse_html_racelist
from .base import BaseScraper

_logger: Logger = getLogger(__name__)

class RaceInfoScraper(BaseScraper):
    """To get race information
    """

    __url_format = f"{BOATRACEJP_BASE_URL}/racelist?rno={{race}}&jcd={{stadium:02d}}&hd={{date}}"  # noqa

    def __init__(
        self,
        driver: webdriver.remote.webdriver.WebDriver,
        logger: Logger = _logger,
    ):
        super().__init__(driver, parse_html_racelist, logger)

    @classmethod
    def make_url(cls, d: datetime.date, stadium: int, race: int) -> str:
        """Make target URL.

        Args:
            d (datetime.date): race date
            stadium (int): stadium no.
            race (int): race no.

        Returns:
            str: URL
        """
        return cls.__url_format.format(
            stadium=stadium,
            date=d.strftime("%Y%m%d"),
            race=race,
        )

    def get(self, d: datetime.date, stadium: int, race: int) -> Dict[str, Any]:
        """Get race information.

        Args:
            d (datetime.date): race date
            stadium (int): stadium no.
            race (int): race no.

        Raises:
            ValueError: Occurred when invalid date given.
            ValueError: Occurred when invalid stadium no. given.
            ValueError: Occurred when invalid race no. given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        validate_stadium(stadium)
        validate_race(race)
        dic = super().get(d, stadium, race)
        dic.update(
            date=d.strftime("%Y-%m-%d"),
            stadium=stadium,
            race=race,
        )
        return dic



from .parse_html_beforeinfo import parse_html_beforeinfo
from .parse_html_index import parse_html_index
from .parse_html_odds2tf import parse_html_odds2tf
from .parse_html_odds3f import parse_html_odds3f
from .parse_html_odds3t import parse_html_odds3t
from .parse_html_oddsk import parse_html_oddsk
from .parse_html_oddstf import parse_html_oddstf
from .parse_html_raceindex import parse_html_raceindex
from .parse_html_racelist import parse_html_racelist
from .parse_html_raceresult import parse_html_raceresult

__all__ = [
    parse_html_index.__name__,
    parse_html_raceindex.__name__,
    parse_html_racelist.__name__,
    parse_html_oddstf.__name__,
    parse_html_oddsk.__name__,
    parse_html_odds2tf.__name__,
    parse_html_odds3t.__name__,
    parse_html_odds3f.__name__,
    parse_html_beforeinfo.__name__,
    parse_html_raceresult.__name__,
]



from abc import ABCMeta, abstractmethod
from logging import Logger, getLogger
from typing import Any

from selenium import webdriver

from ..exceptions import UnableActionException
from ..user_information import UserInformation

_logger: Logger = getLogger(__name__)


class BaseOperator(metaclass=ABCMeta):
    """Base class for operator
    """

    def __init__(
        self,
        user: UserInformation,
        driver: webdriver.remote.webdriver.WebDriver,
        logger: Logger = _logger,
    ):
        self._user = user
        self._driver = driver
        self._logger = logger

    @abstractmethod
    def do(self, *args, **kwargs) -> Any:
        raise NotImplementedError()


class DriverCheckMixin:

    def _check_driver(self):
        """Check whether self._driver is Chrome, Firefox or Edge

        Raises:
            UnableActionException: Occurred when self._driver is not Chrome, Firefox or Edge
        """  # noqa
        if not (
            isinstance(self._driver, webdriver.Chrome)
            or isinstance(self._driver, webdriver.Firefox)
            or isinstance(self._driver, webdriver.Edge)
        ):
            raise UnableActionException(
                "Given driver is not Chrome, Firefox or Edge. "
                "So the operator cannot do anything."
            )



import time

from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.ui import WebDriverWait

from ..exceptions import (
    InactiveRace,
    InactiveStadium,
    InsufficientDepositException,
    ZeroDepositException,
)
from ..validator import validate_race, validate_stadium
from .base import BaseOperator, DriverCheckMixin
from .static import get_bet_limit, visit_ibmbraceorjp


class BettingOperator(BaseOperator, DriverCheckMixin):
    """To bet.
    """

    def do(
        self,
        stadium: int,
        race: int,
        betdict: dict,
        timeout: int = 15,
        raise_validation_error: bool = False,
    ) -> bool:
        """To bet money on the race.

        Args:
            stadium (int): stadium no.
            race (int): race no.
            betdict (dict): betting target dictionary.
            timeout (int, optional): timeout parameter. Defaults to 15.
            raise_validation_error (bool, optional): whether raise validation error. Defaults to False.

        Raises:
            ValueError: Occurred when raise_validation_error is True and the vote password is not set.
            ValueError: Occurred when invalid stadium no. given.
            ValueError: Occurred when invalid race no. given.
            UnableActionException:
                Occurred when driver is not Chrome, Firefox or Edge.
            ZeroDepositException:
                Occurred when no deposit.
            InactiveStadium:
                Occurred when stadium no. is valid but no races are there.
            InactiveRace:
                Occurred when race no. is valid but the race is probabily over.
            InsufficientDepositException:
                Occurred when the sum of your bet is greater than deposit.

        Returns:
            bool: whether betting is succeeded

        NOTE:
            betdict keys:
                'win',
                'placeshow',
                'exacta',
                'quinella',
                'quinellaplace',
                'trifecta',
                'trio',

        TODO: to create the data structure for betdict.
        """  # noqa
        self._check_driver()
        validate_race(race)
        validate_stadium(stadium)
        return self.__bet(
            stadium,
            race,
            betdict,
            timeout=timeout,
            raise_validation_error=raise_validation_error,
        )

    def __bet(
        self,
        stadium: int,  # TODO rename stadium -> stadium
        race: int,
        betdict: dict,
        timeout: int = 15,
        raise_validation_error: bool = False,
    ) -> bool:
        # validation
        if self._user.vote_pass is None:
            if raise_validation_error:
                raise ValueError('Vote password is not set.')
            self._logger.error('Vote password is not set.')
            return False

        # visit
        visit_ibmbraceorjp(self._user, self._driver, timeout)

        # TODO when limit is not enough
        limit = get_bet_limit(self._user, self._driver, timeout)
        if limit == 0:
            # TODO add test
            raise ZeroDepositException('Current deposit is zero.')

        # click stadium
        WebDriverWait(self._driver, timeout).until(EC.presence_of_element_located((By.ID, f'jyo{stadium:02d}')))  # noqa
        element = self._driver.find_element(By.ID, f'jyo{stadium:02d}')
        if 'borderNone' in element.get_attribute('class'):  # type: ignore  # noqa
            # invalid stadium case
            # TODO add test
            raise InactiveStadium(f'The stadium {stadium:02d} has not active races')  # noqa
        else:
            element.click()

        # click race
        WebDriverWait(self._driver, timeout).until(EC.presence_of_element_located((By.ID, f'selRaceNo{race:02d}')))  # noqa
        element = self._driver.find_element(By.ID, f'selRaceNo{race:02d}')
        if 'end' in element.get_attribute('class'):  # type: ignore  # noqa
            # invalid race case
            # TODO add test
            raise InactiveRace(f'Race{race:02d} in stadium {stadium:02d} has ended or is not hold.')  # noqa
        else:
            element.click()

        # create betting list
        # TODO make kinds constant
        amount: int = 0
        for kind_idx, kind in enumerate([
            'win',
            'placeshow',
            'exacta',
            'quinella',
            'quinellaplace',
            'trifecta',
            'trio',
        ]):
            bet_dict_for_kind = betdict.get(kind, None)

            # if not given
            if bet_dict_for_kind is None:
                self._logger.info(f'Skip betting {kind}')
                continue

            # click kind
            self._driver.find_element(By.ID, f'betkati{kind_idx+1}').click()
            time.sleep(1)

            # input bet
            for order, amt in bet_dict_for_kind.items():
                # TODO make sep constant
                sep = '=' if '=' in order else '-'
                boats = tuple(map(int, order.split(sep)))
                print(kind, order, amt)
                for boat_idx, boat in enumerate(boats):
                    self._driver.find_element(By.ID, f'regbtn_{boat}_{boat_idx+1}').click()  # noqa

                self._driver.find_element(By.ID, 'amount').send_keys('\b'*10)  # noqa
                self._driver.find_element(By.ID, 'amount').send_keys(amt//100)  # noqa
                self._driver.find_element(By.ID, 'regAmountBtn').click()

                amount = amount + amt

        # complete input
        self._driver.find_element(By.CLASS_NAME, 'btnSubmit').click()

        # insufficient depost
        if amount > limit:
            # TODO add test
            raise InsufficientDepositException(
                f'Your betting amount is {amount}, '
                f'but your current deposit is {limit}.'
            )

        # confirmation
        self._driver.find_element(By.ID, 'amount').send_keys(str(amount))
        self._driver.find_element(By.ID, 'pass').send_keys(self._user.vote_pass)  # noqa
        self._driver.find_element(By.ID, 'submitBet').click()
        self._driver.find_element(By.ID, 'ok').click()
        # TODO check whether amount is equal to the amount betted
        # TODO vote time limit comes during this function

        return True



import os

import pytest
from selenium.common.exceptions import WebDriverException

from ._driver_fixutures import driver  # noqa

EXPECTED_DIREC = 'tests/data'


@pytest.mark.integrate
def test_driver_get(driver):  # noqa
    # preparation
    url = 'https://example.com'
    # expected
    path = os.path.join(EXPECTED_DIREC, "expected_example.com.html")
    with open(path, 'r', encoding='utf-8-sig') as f:
        expected = f.read()
    expected = ''.join(expected.split())\
                 .replace("<!doctypehtml>", '')\
                 .replace("/>", '>')
    # actual
    driver.get(url)
    actual = driver.page_source
    actual = ''.join(actual.split())\
               .replace("<!doctypehtml>", '')\
               .replace("/>", '>')
    # assert
    assert actual == expected


@pytest.mark.integrate
def test_driver_get_404_not_found(driver):  # noqa
    # preparation
    url = 'https://this_does_not_exists_hogehogehogehoge'
    # test
    with pytest.raises(WebDriverException):
        driver.get(url)



import datetime
from logging import Logger, getLogger
from typing import Any, Dict, Optional

from selenium import webdriver

from . import operator, scraper
from .drivers import create_httpget_driver
from .exceptions import UserInformationNotGivenException
from .user_information import UserInformation
from .validator import validate_date, validate_race, validate_stadium

_logger: Logger = getLogger(__name__)


class PyJPBoatrace(object):
    """Python-based Japanese Boatrace Tools Class

    Attributes:
        Stadiums (scraper.StadiumsScraper)
        Races (scraper.RacesScraper)
        RaceInfo (scraper.RaceInfoScraper)
        JustBeforeInfo (scraper.JustBeforeInfoScraper)
        WinPlaceshowOdds (scraper.WinPlaceshowOddsScraper)
        QuinellaplaceOdds (scraper.QuinellaplaceOddsScraper)
        ExactaQuinellaOdds (scraper.ExactaQuinellaOddsScraper)
        TrioOdds (scraper.TrioOddsScraper)
        TrifectaOdds (scraper.TrifectaOddsScraper)
        Result (scraper.ResultScraper)

        Bet (operator.BettingOperator)
        Depost (operator.DepositOperator)
        Withdraw (operator.WithdrawOperator)
        BettingLimitCheck (operator.BettingLimitCheckOperator)

    """

    def __init__(
        self,
        driver: webdriver.remote.webdriver.WebDriver = create_httpget_driver(),  # type: ignore  # noqa
        user_information: Optional[UserInformation] = None,
        close_driver_when_closing_pyjpboatrace: bool = True,
        logger: Logger = _logger,
    ):
        """Python-based Japanese Boatrace Tools Class

        Args:
            driver (webdriver.remote.webdriver.WebDriver, optional): webdriver. Defaults to create_httpget_driver().
            user_information (Optional[UserInformation], optional): user information. Defaults to None.
            close_driver_when_closing_pyjpboatrace (bool, optional): If True, close driver when calling self.close(). Defaults to True.
            logger (Logger, optional): logger. Defaults to getLogger(__name__).
        """  # noqa
        self.__driver = driver
        self.__user_information = user_information
        self.__close_driver_when_closing_pyjpboatrace = close_driver_when_closing_pyjpboatrace  # noqa
        self.__logger = logger

        self.Stadiums = scraper.StadiumsScraper(driver)
        self.Races = scraper.RacesScraper(driver)
        self.RaceInfo = scraper.RaceInfoScraper(driver)
        self.JustBeforeInfo = scraper.JustBeforeInfoScraper(driver)
        self.WinPlaceshowOdds = scraper.WinPlaceshowOddsScraper(driver)
        self.QuinellaplaceOdds = scraper.QuinellaplaceOddsScraper(driver)
        self.ExactaQuinellaOdds = scraper.ExactaQuinellaOddsScraper(driver)
        self.TrioOdds = scraper.TrioOddsScraper(driver)
        self.TrifectaOdds = scraper.TrifectaOddsScraper(driver)
        self.Result = scraper.ResultScraper(driver)

        self.Bet: Optional[operator.BettingOperator]
        self.Depost: Optional[operator.DepositOperator]
        self.Withdraw: Optional[operator.WithdrawOperator]
        self.BettingLimitCheck: Optional[operator.BettingLimitCheckOperator]
        if user_information is not None:
            self.Bet = operator.BettingOperator(user_information, driver)
            self.Depost = operator.DepositOperator(user_information, driver)
            self.Withdraw = operator.WithdrawOperator(user_information, driver)
            self.BettingLimitCheck = operator.BettingLimitCheckOperator(user_information, driver)  # noqa
        else:
            self.Bet = None
            self.Depost = None
            self.Withdraw = None
            self.BettingLimitCheck = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if self.__close_driver_when_closing_pyjpboatrace:
            self.__driver.close()
            self.__logger.info("Closed the driver")

    def get_stadiums(self, d: datetime.date) -> Dict[str, Any]:
        """Get stadiums where races are held on the given day.

        Args:
            d (datetime.date): race date

        Raises:
            ValueError: Occurred when invalid date given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        return self.Stadiums.get(d)

    def get_12races(self, d: datetime.date, stadium: int) -> Dict[str, Any]:
        """Get 12 races held in the given stadium on the given date.

        Args:
            d (datetime.date): race date
            stadium (int): stadium no.

        Raises:
            ValueError: Occurred when invalid date given.
            ValueError: Occurred when invalid stadium no. given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        validate_stadium(stadium)
        return self.Races.get(d, stadium)

    def get_race_info(self, d: datetime.date, stadium: int, race: int) -> Dict[str, Any]:  # noqa
        """Get race information.

        Args:
            d (datetime.date): race date
            stadium (int): stadium no.
            race (int): race no.

        Raises:
            ValueError: Occurred when invalid date given.
            ValueError: Occurred when invalid stadium no. given.
            ValueError: Occurred when invalid race no. given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        validate_stadium(stadium)
        validate_race(race)
        return self.RaceInfo.get(d, stadium, race)

    def get_odds_win_placeshow(
        self,
        d: datetime.date,
        stadium: int,
        race: int
    ) -> Dict[str, Any]:
        """Get win/placeshow odds

        Args:
            d (datetime.date): race date
            stadium (int): stadium no.
            race (int): race no.

        Raises:
            ValueError: Occurred when invalid date given.
            ValueError: Occurred when invalid stadium no. given.
            ValueError: Occurred when invalid race no. given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        validate_stadium(stadium)
        validate_race(race)
        return self.WinPlaceshowOdds.get(d, stadium, race)

    def get_odds_quinellaplace(
        self,
        d: datetime.date,
        stadium: int,
        race: int
    ) -> Dict[str, Any]:
        """Get quinellaplace odds

        Args:
            d (datetime.date): race date
            stadium (int): stadium no.
            race (int): race no.

        Raises:
            ValueError: Occurred when invalid date given.
            ValueError: Occurred when invalid stadium no. given.
            ValueError: Occurred when invalid race no. given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        validate_stadium(stadium)
        validate_race(race)
        return self.QuinellaplaceOdds.get(d, stadium, race)

    def get_odds_exacta_quinella(
        self,
        d: datetime.date,
        stadium: int,
        race: int
    ) -> Dict[str, Any]:
        """Get exacta/quinella odds

        Args:
            d (datetime.date): race date
            stadium (int): stadium no.
            race (int): race no.

        Raises:
            ValueError: Occurred when invalid date given.
            ValueError: Occurred when invalid stadium no. given.
            ValueError: Occurred when invalid race no. given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        validate_stadium(stadium)
        validate_race(race)
        return self.ExactaQuinellaOdds.get(d, stadium, race)

    def get_odds_trifecta(
        self,
        d: datetime.date,
        stadium: int,
        race: int
    ) -> Dict[str, Any]:
        """Get trifecta odds

        Args:
            d (datetime.date): race date
            stadium (int): stadium no.
            race (int): race no.

        Raises:
            ValueError: Occurred when invalid date given.
            ValueError: Occurred when invalid stadium no. given.
            ValueError: Occurred when invalid race no. given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        validate_stadium(stadium)
        validate_race(race)
        return self.TrifectaOdds.get(d, stadium, race)

    def get_odds_trio(
        self,
        d: datetime.date,
        stadium: int,
        race: int
    ) -> Dict[str, Any]:
        """Get trio odds

        Args:
            d (datetime.date): race date
            stadium (int): stadium no.
            race (int): race no.

        Raises:
            ValueError: Occurred when invalid date given.
            ValueError: Occurred when invalid stadium no. given.
            ValueError: Occurred when invalid race no. given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        validate_stadium(stadium)
        validate_race(race)
        return self.TrioOdds.get(d, stadium, race)

    def get_just_before_info(
        self,
        d: datetime.date,
        stadium: int,
        race: int
    ) -> Dict[str, Any]:
        """Get just-before race information

        Args:
            d (datetime.date): race date
            stadium (int): stadium no.
            race (int): race no.

        Raises:
            ValueError: Occurred when invalid date given.
            ValueError: Occurred when invalid stadium no. given.
            ValueError: Occurred when invalid race no. given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        validate_stadium(stadium)
        validate_race(race)
        return self.JustBeforeInfo.get(d, stadium, race)

    def get_race_result(
        self,
        d: datetime.date,
        stadium: int,
        race: int
    ) -> Dict[str, Any]:
        """Get race result.

        Args:
            d (datetime.date): race date
            stadium (int): stadium no.
            race (int): race no.

        Raises:
            ValueError: Occurred when invalid date given.
            ValueError: Occurred when invalid stadium no. given.
            ValueError: Occurred when invalid race no. given.

        Returns:
            Dict[str, Any]: scraped data
        """
        validate_date(d)
        validate_stadium(stadium)
        validate_race(race)
        return self.Result.get(d, stadium, race)

    def deposit(self, num_of_thousands_yen: int) -> None:
        """To deposit money.

        Args:
            depo_amt_unit_thousands_yen (int): the amount of deposit.

        Raises:
            UnableActionException:
                Occurred when driver is not Chrome, Firefox or Edge.
        """
        # deposit
        if self.Depost is not None:
            self.Depost.do(num_of_thousands_yen)
        else:
            raise UserInformationNotGivenException()

    def get_bet_limit(self) -> int:
        """To check the amount of money deposited.

        Raises:
            UnableActionException:
                Occurred when driver is not Chrome, Firefox or Edge.

        Returns:
            int: the amount of deposited money
        """
        # get bet limit
        if self.BettingLimitCheck is not None:
            return self.BettingLimitCheck.do()
        else:
            raise UserInformationNotGivenException()

    def withdraw(self) -> None:
        """To withdraw deposit.

        Raises:
            ZeroDepositException:
                Occurred when no deposit.
            UnableActionException:
                Occurred when driver is not Chrome, Firefox or Edge.
        """
        # withdraw
        if self.Withdraw is not None:
            self.Withdraw.do()
        else:
            raise UserInformationNotGivenException()

    def bet(
        self,
        stadium: int,
        race: int,
        trifecta_betting_dict: Optional[Dict[str, int]] = None,
        trio_betting_dict: Optional[Dict[str, int]] = None,
        exacta_betting_dict: Optional[Dict[str, int]] = None,
        quinella_betting_dict: Optional[Dict[str, int]] = None,
        quinellaplace_betting_dict: Optional[Dict[str, int]] = None,
        win_betting_dict: Optional[Dict[str, int]] = None,
        placeshow_betting_dict: Optional[Dict[str, int]] = None,
    ) -> bool:
        """To bet money on the race.

        Args:
            stadium (int): stadium no.
            race (int): race no.
            trifecta_betting_dict (Dict[str, int], optional):
                Betting dictionary for trifecta.
                Defaults to None, which means no bet.
                e.g. {"1-2-3": 100, }
            trio_betting_dict (Dict[str, int], optional):
                Betting dictionary for trio.
                Defaults to None, which means no bet.
                e.g. {"1=2=3": 100, }
            exacta_betting_dict (Dict[str, int], optional):
                Betting dictionary for exacta.
                Defaults to None, which means no bet.
                e.g. {"1-2": 100, }
            quinella_betting_dict (Dict[str, int], optional):
                Betting dictionary for quinella.
                Defaults to None, which means no bet.
                e.g. {"1=2": 100, }
            quinellaplace_betting_dict (Dict[str, int], optional):
                Betting dictionary for quinellaplace.
                Defaults to None, which means no bet.
                e.g. {"1=3": 100, }
            win_betting_dict (Dict[str, int], optional):
                Betting dictionary for win.
                Defaults to None, which means no bet.
                e.g. {"1": 100, }
            placeshow_betting_dict (Dict[str, int], optional):
                Betting dictionary for placeshow.
                Defaults to None, which means no bet.
                e.g. {"1": 100, }

        Raises:
            ValueError: Occurred when invalid stadium no. given.
            ValueError: Occurred when invalid race no. given.
            UnableActionException:
                Occurred when driver is not Chrome, Firefox or Edge.
            ZeroDepositException:
                Occurred when no deposit.
            InactiveStadium:
                Occurred when stadium no. is valid but no races are there.
            InactiveRace:
                Occurred when race no. is valid but the race is probabily over.
            InsufficientDepositException:
                Occurred when the sum of your bet is greater than deposit.
            UserInformationNotGivenException:
                Occurred when user_information is not given when instantiating PyJPBoatrace.

        Returns:
            bool: whether betting is succeeded

        NOTE:
            betdict keys:
                'win',
                'placeshow',
                'exacta',
                'quinella',
                'quinellaplace',
                'trifecta',
                'trio',

        TODO: to create the data structure for betting dict.
        """  # noqa
        # preprocess
        trifecta_betting_dict = trifecta_betting_dict or {}
        trio_betting_dict = trio_betting_dict or {}
        exacta_betting_dict = exacta_betting_dict or {}
        quinella_betting_dict = quinella_betting_dict or {}
        quinellaplace_betting_dict = quinellaplace_betting_dict or {}
        win_betting_dict = win_betting_dict or {}
        placeshow_betting_dict = placeshow_betting_dict or {}

        # create bet dict
        betdict = {
            'trifecta': trifecta_betting_dict,
            'trio': trio_betting_dict,
            'exacta': exacta_betting_dict,
            'quinella': quinella_betting_dict,
            'quinellaplace': quinellaplace_betting_dict,
            'win': win_betting_dict,
            'placeshow': placeshow_betting_dict,
        }

        # bet
        if self.Bet is None:
            raise UserInformationNotGivenException()
        return self.Bet.do(
            stadium=stadium,
            race=race,
            betdict=betdict
        )




def str2num(s: str, typ: type, default_val=None):
    ''' string to number whose type is the type given as typ.

    typ must be int, float or complex

    Note: Failure of casting returns default_val
    '''
    if typ not in [int, float, complex]:
        raise NotImplementedError(
            f'typ must be int, float or complex, but {typ.__name__} given.'
        )

    try:
        return typ(s)
    except ValueError:
        return default_val



from .base import BaseScraper
from .exacta_quinella_odds_scraper import ExactaQuinellaOddsScraper
from .just_before_info_scraper import JustBeforeInfoScraper
from .quinellaplace_odds_scraper import QuinellaplaceOddsScraper
from .race_info_scraper import RaceInfoScraper
from .races_scraper import RacesScraper
from .result_scraper import ResultScraper
from .stadiums_scraper import StadiumsScraper
from .trifecta_odds_scraper import TrifectaOddsScraper
from .trio_odds_scraper import TrioOddsScraper
from .win_placeshow_odds_scraper import WinPlaceshowOddsScraper

__all__ = [
    BaseScraper.__name__,
    StadiumsScraper.__name__,
    RacesScraper.__name__,
    RaceInfoScraper.__name__,
    JustBeforeInfoScraper.__name__,
    WinPlaceshowOddsScraper.__name__,
    QuinellaplaceOddsScraper.__name__,
    ExactaQuinellaOddsScraper.__name__,
    TrioOddsScraper.__name__,
    TrifectaOddsScraper.__name__,
    ResultScraper.__name__,
]



class BasePyJPBoatraceException(Exception):
    """Base class for all pyjpboatrace-specific exceptions."""
    pass


class NoDataException(BasePyJPBoatraceException):
    pass


class RaceCancelledException(BasePyJPBoatraceException):
    def __init__(self, msg=None):
        super().__init__(msg)


class UnableActionException(BasePyJPBoatraceException):
    def __init__(self, msg=None):
        super().__init__(msg)


class LoginFailException(BasePyJPBoatraceException):
    pass


class InsufficientDepositException(BasePyJPBoatraceException):
    pass


class ZeroDepositException(InsufficientDepositException):
    pass


class VoteNotInTimeException(BasePyJPBoatraceException):
    pass


class InactiveStadium(BasePyJPBoatraceException):
    pass


class InactiveRace(BasePyJPBoatraceException):
    pass


class UserInformationNotGivenException(BasePyJPBoatraceException):
    pass


class UnexpectedException(BasePyJPBoatraceException):
    pass


__all__ = [
    "BasePyJPBoatraceException",
    "NoDataException",
    "RaceCancelledException",
    "UnableActionException",
    "LoginFailException",
    "InsufficientDepositException",
    "ZeroDepositException",
    "VoteNotInTimeException",
    "InactiveStadium",
    "InactiveRace",
    "UserInformationNotGivenException",
    "UnexpectedException",
]
